/*===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 * 
 * IBM SDK, Java(tm) Technology Edition, v7
 * (C) Copyright IBM Corp. 2014, 2014. All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *===========================================================================
 */
/*
 * Copyright (c) 1995, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class defines the RDMA network.
 * 
 * This package private class provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.IOException;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import java.util.List;
import java.io.File;
import java.util.Scanner;

class RDMANetworkProvider implements NetworkProvider {
	// native methods
	private native void initialize0(boolean zCopy, int zCopyThreshold,
								int sbufSize, int rbufSize, int sqSize, int rqSize, int sqInline) throws IOException;
	private native int cleanup0();
	
	// this provider name
	private static final String name = "RDMA";

	// Preferred addresses
	private InetAddress preferredAddress = null;
	private static final boolean zeroCopy;
	private static final int zeroCopyThreshold;
	private static final int DFLT_ZERO_COPY_THRESHOLD = 256; // KB
	private static final int MIN_ZERO_COPY_THRESHOLD = 64; // KB
	private static final int MAX_ZERO_COPY_THRESHOLD = 16384; // KB
	private static final boolean debugInterception;

	private static final String SEND_BUFFER_SIZE = "SendBufferSize";
	private static final String RECEIVE_BUFFER_SIZE = "ReceiveBufferSize";
	private static final String SEND_QUEUE_SIZE = "SendQueueSize";
	private static final String RECEIVE_QUEUE_SIZE = "ReceiveQueueSize";
	private static final String SEND_INLINE_SIZE = "SendInlineSize";
	
	private static final int sendBufferSize;
	private static final int receiveBufferSize;
	private static final int sendQueueSize;
	private static final int receiveQueueSize;
	private static final int sendInlineSize;

	static {
		String zeroCopyPropStr = null;
		String zeroCopyThresholdPropStr = null;
		boolean copy = false;
		int copyThreshold = DFLT_ZERO_COPY_THRESHOLD;
		boolean debug = false;
		String debugPropStr = null;
		String sockOptsPropFile = null;
		int sbufSize = 0;
		int rbufSize = 0;
		int sqSize = 0;
		int rqSize = 0;
		int sqInline = 0;

		java.security.AccessController.doPrivileged(
				new sun.security.action.LoadLibraryAction("jsor"));

		zeroCopyPropStr = AccessController.doPrivileged(new GetPropertyAction("com.ibm.net.rdma.zeroCopy"));
		if (zeroCopyPropStr != null) {
			copy = Boolean.parseBoolean(zeroCopyPropStr);
		}
		zeroCopy = copy;

		zeroCopyThresholdPropStr = AccessController.doPrivileged(new GetPropertyAction("com.ibm.net.rdma.zeroCopyThreshold"));
		if (zeroCopyThresholdPropStr != null) {
			try {
				copyThreshold = Integer.parseInt(zeroCopyThresholdPropStr);
				if (copyThreshold < MIN_ZERO_COPY_THRESHOLD) {
					System.out.format("[JSOR Warning] Invalid zero copy threshold \'%d\' - defaulting to %d KB%n",
							copyThreshold, MIN_ZERO_COPY_THRESHOLD);
					copyThreshold = MIN_ZERO_COPY_THRESHOLD;
				} else if (copyThreshold > MAX_ZERO_COPY_THRESHOLD) {
					System.out.format("[JSOR Warning] Invalid zero copy threshold \'%d\' - defaulting to %d KB%n",
							copyThreshold, MAX_ZERO_COPY_THRESHOLD);
					copyThreshold = MAX_ZERO_COPY_THRESHOLD;
				}
			} catch (NumberFormatException e) {
				System.out.println("[JSOR Warning] Invalid \"com.ibm.net.rdma.zeroCopyThreshold\" property value ("
						+e.getMessage()+"): defaulting to "+DFLT_ZERO_COPY_THRESHOLD+" KB");
				copyThreshold = DFLT_ZERO_COPY_THRESHOLD;
			}
		}
		zeroCopyThreshold = copyThreshold;

		debugPropStr = AccessController.doPrivileged(new GetPropertyAction("com.ibm.net.rdma.debug"));
		if (debugPropStr != null) {
			debug = Boolean.parseBoolean(debugPropStr);
		}
		debugInterception = debug;

		sockOptsPropFile = AccessController.doPrivileged(new GetPropertyAction("com.ibm.net.rdma.socketOptions"));
		if (sockOptsPropFile != null) {
			Scanner scanner = null;
		
			try {
				scanner = new Scanner(new File(sockOptsPropFile));
			} catch(Exception e) {
				// Ignore file not found issues
			}

			while ((scanner != null) && scanner.hasNextLine()) {
				String line = scanner.nextLine().trim();
				// empty line
				if (line.length() == 0) {
					continue;
				}
				// process only lines having <key>=<value> format
				String[] s = line.split("=");
				if (s.length != 2) {
					continue;
				}
				String key = s[0].trim();
				String value = s[1].trim();
				/* no validation happens here
 				 * rdma runtime will take care of it later during
 				 * initialization time
 				 */
				try {
					switch(key) {
					case SEND_BUFFER_SIZE:
						sbufSize = Integer.parseInt(value);
						break;
					case RECEIVE_BUFFER_SIZE:
						rbufSize = Integer.parseInt(value);
						break;
					case SEND_QUEUE_SIZE:
						sqSize = Integer.parseInt(value);
						break;
					case RECEIVE_QUEUE_SIZE:
						rqSize = Integer.parseInt(value);
						break;
					case SEND_INLINE_SIZE:
						sqInline = Integer.parseInt(value);
						break;
					default:
						break;
					}
				} catch (Exception e) {
					// ignore parsing errors
				}
			}
		} 
		sendBufferSize = sbufSize;
		receiveBufferSize = rbufSize;
		sendQueueSize = sqSize;
		receiveQueueSize = rqSize;
		sendInlineSize = sqInline;
	}

	/**
	 * initialize the RDMA network
	 */
	public void initialize() throws IOException {
		try {
			initialize0(zeroCopy, zeroCopyThreshold, sendBufferSize, receiveBufferSize, sendQueueSize, receiveQueueSize, sendInlineSize);
			if (debugInterception) {
				System.out.println("[JSOR Info] RDMA network provider inialized");
			}
		} catch (Exception e) {
			throw new IOException(e.getMessage());
		}
	}
	
	/**
	 * destroy RDMA handles that are assigned during
	 * network initialization
	 */
	public void cleanup() {
		cleanup0();
	}
	
	/**
	 * create an RDMA socket impl
	 */
	public SocketImpl createImpl(SocketImpl oldImpl) {
		SocketImpl impl = new RDMASocketImpl(oldImpl, zeroCopy, zeroCopyThreshold, debugInterception);
		if (debugInterception) {
			System.out.println("[JSOR Info] Socket switched to RDMA mode");
		}
		return impl;
	}

	/**
	 * do post-accept operations for an RDMA connection
	 */
	public void postAccept(SocketImpl serverSocketImpl, SocketImpl acceptedSocketImpl) throws IOException {
		// Do any post accept ops that are necessary in RDMA mode
		// Right now, they are none to perform
	}

	/**
	 * get the name of this network provider 
	 */
	public String getName() {
		return name;
	}

	/**
	 * Set prefered address based on user preference for RDMA transfer
	 * Preferred address could be one of the valid ib or eth interface addresses
	 * By default we pickup the first available ib interface address if none is specified
	 * Don't rely on eth interface addresses for setting the default because we don't
	 * have knowledge about their RDMA capabilities
	 */
	public void setPreferredAddress(List<String> ibAddresses, List<String> ethAddresses) {
		InetAddress prefAddr = null;
		try {
			String preferredAddressStr = AccessController.doPrivileged(new GetPropertyAction("com.ibm.net.rdma.preferredAddress"));
			// preferred address specified is one of the avail IB  or Ethernet (ROCE) addresses
			if (preferredAddressStr != null &&
					((!ibAddresses.isEmpty() && ibAddresses.contains(preferredAddressStr)) ||
					(!ethAddresses.isEmpty() && ethAddresses.contains(preferredAddressStr)))) {
				prefAddr = InetAddress.getByName(preferredAddressStr);
			} else {
				// by default we pickup the first avail IB addr
				if (!ibAddresses.isEmpty()) {
					prefAddr = InetAddress.getByName(ibAddresses.get(0));
				}
			}
		} catch (IOException e) {
			// let the preferredAddress be null
 		}
 		preferredAddress = prefAddr;
	}

 	/**
	 * Get prefered addresses for RDMA transfer
	 */
 	public InetAddress getPreferredAddress() {
		return preferredAddress;
	}
}
